from __future__ import annotations

import inspect
import math
import os
import random
import re
import shlex
import string
import subprocess
from collections.abc import Callable, Iterator
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from re import Pattern
from textwrap import dedent
from typing import Any, Final

import yaml

from jett.__types import DictData
from jett.errors import CmdExecError


def join_newline(value: list[str]) -> str:
    """Join the element in list of string with newline. This function use to
    avoid error when joining inside a format string syntax like
    `f\"{''.join(['foo', 'bar'])}\"`.

    Args:
        value: list[str]
    """
    return "\n".join(value)


def load_yaml(path: str | Path, add_info: bool = True) -> DictData:
    """Load YAML data. It will return empty data when the path does not exist.

    Args:
        path: A YAML filepath.
        add_info:

    Returns:
        DictData: return configuration data that extract from YAML file or
        return empty dict object if it does not exist.
    """
    path: Path = Path(path)
    if path.exists():
        file_state: os.stat = path.lstat()
        return (
            {
                "created_at": file_state.st_ctime,
                "updated_at": file_state.st_mtime,
                "parent_dir": path.parent,
            }
            if add_info
            else {}
        ) | yaml.safe_load(path.open(mode="r"))
    return {}


def write_yaml(path: str | Path, data: Any) -> None:
    """Write YAML file with an input data."""
    with open(path, mode="w", encoding="utf-8") as f:
        yaml.dump(data, f, default_flow_style=False, sort_keys=False)


def get_dt_now() -> datetime:
    """Get the current datetime object with UTC timezone.

    Returns:
        datetime: The current datetime object with UTC timezone.
    """
    return datetime.now(timezone.utc)


def get_dt_latest() -> datetime:
    return datetime(9999, 12, 31)


def char_to_snake_case(name: str) -> str:
    """Converts a string to snake_case.

    Examples:
        - 'columnName'  -> 'column_name'
        - 'Column Name' -> 'column_name'
        - 'ColumnID'    -> 'column_id'
        - 'IDColumn'    -> 'id_column'
        - 'ID_Column_Name' -> 'id_column_name'
    """
    segments = name.split("_")
    converted_segments = []

    for segment in segments:
        if not segment:  # Handle multiple consecutive underscores
            continue

        # Apply camelCase/PascalCase conversion to each segment
        pattern: re.Pattern[str] = re.compile(
            r"[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+"
        )
        words: list[str] = pattern.findall(segment)
        if words:
            converted_segments.extend(w.lower() for w in words)

    return "_".join(converted_segments)


def to_snake_case(name: str) -> str:
    """Converts a string to snake_case, handling struct-like syntax.

    Examples:
        - 'columnName'  -> 'column_name'
        - 'Column Name' -> 'column_name'
        - 'ColumnID'    -> 'column_id'
        - 'IDColumn'    -> 'id_column'
        - 'struct<ACTUAL_TIMESTAMP:timestamp>' -> 'struct<actual_timestamp:timestamp>'
    """

    # NOTE: Handle struct-like syntax by processing field names within angle
    #   brackets
    def convert_struct_fields(match) -> str:
        struct_content = match.group(1)
        # Split by comma to handle individual fields
        fields = struct_content.split(",")
        converted_fields = []

        for field in fields:
            if ":" in field:
                # Split field name from type
                field_name, field_type = field.split(":", 1)
                # Convert only the field name to snake_case
                converted_name = char_to_snake_case(field_name.strip())
                converted_fields.append(f"{converted_name}:{field_type}")
            else:
                # No type specified, just convert the whole field
                converted_fields.append(char_to_snake_case(field.strip()))

        return f"struct<{','.join(converted_fields)}>"

    # NOTE: Check if this is a struct-like pattern
    struct_pattern = re.compile(r"struct<([^>]+)>")
    if struct_pattern.search(name):
        return struct_pattern.sub(convert_struct_fields, name)

    # NOTE: Use original logic for simple strings
    return char_to_snake_case(name)


def is_snake_case(input_str: str) -> bool:
    """Check that string is a snake case with number."""
    return bool(re.match(r"^[a-z0-9_]*$", input_str))


def is_optional(func: Callable, *, key: str) -> bool:
    signature = inspect.signature(func)
    for name, parameter in signature.parameters.items():
        if name == key:
            return parameter.kind == inspect.Parameter.KEYWORD_ONLY
    return False


is_optional_engine: partial[bool] = partial(is_optional, key="engine")


def bool2str(value: bool) -> str:
    """Convert boolean value to string value in (true or false)."""
    return "true" if value else "false"


def exec_command(
    cmd: str | list[str],
    encoding: str = "utf-8",
    **kwargs,
) -> subprocess.Popen[str]:
    """Execution command function that start to run subprocess built-in function
    and catch the stdout and stderr to return result.

    Args:
        cmd (str | list[str]): A command statement or list of command that want
            to execute with the subprocess.
        encoding (str, default "utf-8"): An encoding that use to parse std
            result.
    """
    if isinstance(cmd, str):
        args: list[str] = shlex.split(cmd)
    elif not isinstance(cmd, list) or not all(isinstance(c, str) for c in cmd):
        raise TypeError("The 'args' parameter must be a list of strings.")
    else:
        args: list[str] = cmd
    return subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        encoding=encoding,
        errors="replace",  # Prevent UnicodeDecodeError from crashing the stream
        **kwargs,
    )


def handle_command(
    cmd: str | list[str],
    *,
    encoding: str = "utf-8",
    **kwargs,
) -> Iterator[str]:
    """Executes a command and yields its stdout line-by-line in real-time.

    This function is a generator that provides a secure and robust way to
    run an external command and process its output as a stream. It is designed
    to be resilient, handling errors such as the command not being found,
    permission issues, and non-zero exit codes from the command itself.

    Args:
        cmd: The command and its arguments as a list of strings. This is the
              only supported method for passing arguments to ensure security
              against command injection.
        encoding: The character encoding to use for decoding the subprocess's
                  stdout and stderr streams. Defaults to 'utf-8'.
        **kwargs: Additional keyword arguments to be passed directly to the
                  subprocess.Popen constructor.

    Yields:
        str: A line of output from the command's stdout, with leading/trailing
             whitespace removed.

    Raises:
        FileNotFoundError: If the command executable specified in `args`
           is not found in the system's PATH.
        PermissionError: If the script lacks the necessary permissions to
         execute the command.
        CmdExecutionError: If the command executes but returns a non-zero
           exit code, indicating a failure. The exception instance contains the
           return code and any captured stderr output.
        TypeError: If `args` is not a list of strings.
    """
    if isinstance(cmd, str):
        args: list[str] = shlex.split(cmd)
    elif not isinstance(cmd, list) or not all(isinstance(c, str) for c in cmd):
        raise TypeError("The 'args' parameter must be a list of strings.")
    else:
        args: list[str] = cmd

    proc: subprocess.Popen | None = None
    try:
        proc: subprocess.Popen = exec_command(args, encoding=encoding, **kwargs)

        # The 'iter(proc.stdout.readline, "")' pattern is a robust way to
        # read from a stream until it's closed. It avoids issues with
        # different line endings and is efficient.
        for line in iter(proc.stdout.readline, ""):
            yield line.strip()

        # After stdout is exhausted, wait for the process to terminate.
        # This is crucial to get the final return code.
        return_code = proc.wait()

        # Capture any remaining stderr output for context in case of an error.
        stderr_output = proc.stderr.read()

        if return_code != 0:
            # The command ran but failed. Raise a custom, informative exception.
            command_str = " ".join(map(shlex.quote, args))
            raise CmdExecError(
                f"Command '{command_str}' failed with return code {return_code}",
                return_code=return_code,
                stderr=stderr_output,
            )

    except FileNotFoundError:
        # Provide a more helpful error message for this common issue.
        raise FileNotFoundError(
            f"Command not found: '{args}'. Please ensure it is installed and "
            f"in your system's PATH."
        ) from None  # Suppress the original exception chain for a cleaner message
    except PermissionError:
        raise PermissionError(
            f"Permission denied to execute command: '{args}'."
        ) from None
    finally:
        # This block ensures that resources are cleaned up, even if the
        # generator is not fully consumed (e.g., the consumer breaks the loop).
        if proc:
            # If the process is still running (e.g., generator was closed early),
            # terminate it to prevent orphaned processes.
            if proc.poll() is None:
                try:
                    proc.terminate()  # Ask nicely first
                    proc.wait(timeout=2)  # Give it a moment to die
                except subprocess.TimeoutExpired:
                    proc.kill()  # Forcefully kill if it doesn't terminate
                    proc.wait()  # Reap the killed process

            # Explicitly close the file handles to release OS resources.
            if proc.stdout:
                proc.stdout.close()
            if proc.stderr:
                proc.stderr.close()


def get_random_str(n: int = 10) -> str:
    """Get random string."""
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=n))


def get_random_str_unique(n: int = 10) -> str:
    return f"{get_random_str(n)}{get_dt_now():%Y%m%d%H%M%S}"


def sort_non_sensitive_str(value: list[str]) -> list[str]:
    """Sort list of string by ignoring special characters and case-sensitive."""
    pattern: Pattern[str] = re.compile(r"[^a-zA-Z0-9]")
    return sorted(value, key=lambda s: pattern.sub("", s).lower())


def dt2str(
    val: datetime,
    sep: str = " ",
    timespec: str = "microseconds",
    add_utc_suffix: bool = False,
) -> str:
    """Covert datetime object to str with format `%Y-%M-%d %H:%M:%S.%s`"""
    rs: str = val.isoformat(sep=sep, timespec=timespec)
    return f"{rs}Z" if add_utc_suffix else rs


def clean_string(text: str) -> str:
    """Remove indent, double whitespace, double newlines from string."""
    text = dedent(text)
    text = text.replace("  ", "")
    text = text.split("\n")
    text = list(filter(None, text))
    text = "\n".join(text)
    return text


def format_bytes_humanreadable(num_bytes: int) -> str:
    """Format bytes into human-readable format."""
    if num_bytes == 0:
        return "0 B"

    size_name: tuple[str, ...] = ("B", "KB", "MB", "GB", "TB", "PB")
    i: int = int(math.floor(math.log(num_bytes, 1024)))
    p: float = math.pow(1024, i)
    s = round(num_bytes / p, 2)
    return f"{s} {size_name[i]}"


def truncate_str_with_byte_limit(
    input_str: str, mode: str, max_bytes: int = 1048576
) -> str:
    """Truncate string with maximum size of string cannot greater than given
    maximum bytes support truncate from head or tail.

    Args:
        input_str (str):
        mode (str):
        max_bytes (int, default is 1048576 (1MB)):
    """

    if mode not in ["head", "tail"]:
        raise ValueError("mode must be head or tail")

    encoded_string = input_str.encode("utf-8")
    encoded_string = (
        encoded_string[:max_bytes]
        if mode == "head"
        else encoded_string[-max_bytes:]
    )
    truncated_string = encoded_string.decode("utf-8", "ignore")
    return truncated_string


def regex_by_group(text: str, regex: str, n: int = 0) -> str:
    """Find text with regex and get only first match
    if not found, return empty string
    """
    match = re.search(regex, text)
    if match:
        return match.group(n)
    return ""


def get_exception_group():  # pragma: no cov
    """Return the ExceptionGroup for compatibility with python < 3.11."""
    try:
        from builtins import ExceptionGroup

        return ExceptionGroup
    except ImportError:
        from exceptiongroup import ExceptionGroup

        return ExceptionGroup


def env(name: str, module: str, *, default: str | None) -> str | None:
    return os.getenv(f"JETT__{module.upper()}__{name}", default)


spark_env: partial[str | None] = partial(env, module="SPARK")
tool_env: partial[str] = partial(os.getenv, "JETT_ENV", default="test")

# NOTE: Compiled regex for environment variable substitution is more performant.
#   Syntax: ${{ ENV_VAR:default_value }} (default_value is optional)
ENV_VAR_MATCHER: Final[Pattern[str]] = re.compile(
    r"\$\{\{\s*(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*(?::\s*(?P<default>.*?))?\s*}}"
)


def substitute_env_vars(value: Any) -> Any:
    """Recursively traverses the data structure and substitutes environment
    variables.

    Args:
        value: The current value from the YAML data (can be dict, list, str, etc.).

    Returns:
        The value with environment variables substituted.

    Raises:
        KeyError: If an environment variable is not set and has no default.
    """
    if isinstance(value, dict):
        return {k: substitute_env_vars(v) for k, v in value.items()}
    if isinstance(value, list):
        return [substitute_env_vars(item) for item in value]
    if isinstance(value, str):
        return replace_string_env_var(value)
    return value


def replace_string_env_var(value: str) -> str:
    """Performs the actual environment variable substitution in a string.

    Args:
        value: The string value to process.

    Returns:
        The string with the placeholder replaced by the environment variable's value.
    """
    match = ENV_VAR_MATCHER.match(value)
    if not match:
        return value

    var_name = match.group("name")
    default_value = match.group("default")

    # Use os.getenv for thread-safe access to environment variables
    env_value = os.getenv(var_name)
    if env_value is not None:
        return env_value
    if default_value is not None:
        return default_value

    raise KeyError(
        f"Configuration error: Environment variable '{var_name}' is not set "
        "and no default value was provided."
    )



from __future__ import annotations

from typing import Annotated, Union

from pydantic import Field

from .console import Console
from .restapi import RestAPI

Metric = Annotated[
    Union[
        Console,
        RestAPI,
    ],
    Field(
        discriminator="type",
        description="A metric model registry type.",
    ),
]



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any

from .utils import ENUM_EXTRACT_ARRAY_TYPE

if TYPE_CHECKING:
    from pyspark.sql.types import DataType, StructType

    Changed = dict[str, str | DataType | list[str]]
    ListStr = list[str]
    ListDictAny = list[dict[str, Any]]


def extract_struct_type(
    source_struct: DataType,
    table_struct: DataType,
    parent_cols: list[str] | None = None,
) -> list[Changed]:
    """A recursive function that walks into nested field of PySpark's Struct and
    collect the list of parent columns in each nested level, if the schema
    change found, it calls evaluate_schema_change() to evaluate the changes,
    otherwise, it has nested field.

    Args:
        source_struct: PySpark's schema of source DataFrame
        table_struct: PySpark's schema of table DataFrame
        parent_cols (list[str], default None): A list of parent columns of each
            nested fields

    Returns:
        list[Changed]
    """
    from pyspark.sql.types import ArrayType, StructType

    parent_cols: list[str] = parent_cols or []
    # NOTE: Found nested field or source struct is not the same data type
    #   as table struct e.g. array of string and array of struct.
    if (
        isinstance(source_struct, StructType)
        or source_struct.__class__ != table_struct.__class__
    ):
        return evaluate_schema_change(
            src_schema=source_struct,
            tgt_schema=table_struct,
            parent_cols=parent_cols,
        )

    if isinstance(source_struct, ArrayType):
        parent_cols.append(ENUM_EXTRACT_ARRAY_TYPE)
        return extract_struct_type(
            source_struct=source_struct.elementType,
            table_struct=table_struct.elementType,
            parent_cols=parent_cols,
        )
    return []


def evaluate_schema_change(
    src_schema: StructType,
    tgt_schema: StructType,
    parent_cols: list[str] | None = None,
) -> list[Changed]:
    """Evaluate the schema change (add, alter, and drop) between source schema
    and table schema in each field and nested field.

    Args:
        src_schema: PySpark's schema of source DataFrame
        tgt_schema: PySpark's schema of table DataFrame
        parent_cols: List of parent columns of each nested fields

    Returns:
        list[str]: list of schema changes, see structure of dict in list:
            Drop column
            >>> {
            ...     'type': 'drop_col',
            ...     'source_struct_type': StructField(),
            ...     'parent_cols': [],
            ... }

            Add column
            >>> {
            ...     'type': 'add_col',
            ...     'source_struct_type': StructField(),
            ...     'add_col_after': '',
            ...     'parent_cols': [],
            ... }

            Alter column
            >>> {
            ...     'type': 'alter_col',
            ...     'source_struct_type': StructField(),
            ...     'target_struct_type': StructField(),
            ...     'parent_cols': [],
            ... }
    """
    from pyspark.sql.types import ArrayType, StructField, StructType

    parent_cols: list[str] = parent_cols or []
    diffs: list[Changed] = []

    # NOTE: Handle case the source schema and table schema are not the same type.
    #   This case it will mark `Alter` type.
    if src_schema.__class__ != tgt_schema.__class__:
        diffs.append(
            {
                "type": "alter_col",
                "source_struct_type": src_schema,
                "target_struct_type": tgt_schema,
                "parent_cols": parent_cols,
            }
        )
        return diffs

    ### get to drop columns
    last_drop_col: str | None = None
    to_drop_cols: list[str] = []
    source_cols: list[str] = [s.name for s in src_schema]
    t: StructField
    for t in tgt_schema:
        if t.name not in source_cols:
            diffs.append(
                {
                    "type": "drop_col",
                    "source_struct_type": t,
                    "parent_cols": parent_cols,
                }
            )
            to_drop_cols.append(t.name)
            last_drop_col = t.name

    ### refresh the table schema
    tgt_schema: list[StructField] = [
        t for t in tgt_schema if t.name not in to_drop_cols
    ]

    ### Find new columns, column's data type change, and nested schema change
    nested_schema_changes: list[Changed] = []
    checked_table_cols: list[str] = []
    last_col_added = None
    _table_cols = [t.name for t in tgt_schema]

    for i, s in enumerate(src_schema):
        s: StructType | StructField
        for t in tgt_schema:
            if s.name == t.name:
                if s.dataType != t.dataType:
                    # NOTE: ignore s.nullable != t.nullable and
                    #   s.metadata != t.metadata.
                    if type(s.dataType) not in (StructType, ArrayType):
                        diffs.append(
                            {
                                "type": "alter_col",
                                "source_struct_type": s,
                                "target_struct_type": t,
                                "parent_cols": parent_cols,
                            }
                        )
                    else:
                        # nested schema is not equal
                        nested_schema_changes.append(
                            {
                                "type": "nested_schema_change",
                                "source_struct_type": s,
                                "table_struct_type": t,
                            }
                        )

                # always break if the column name is matched
                last_col_added = t.name
                checked_table_cols.append(t.name)
                break

            else:
                if s.name not in _table_cols and s.name not in [
                    d.get("source_struct_type").name for d in diffs
                ]:
                    # if first index, no add col after, set to None
                    _add_col_after = (
                        last_col_added
                        if (last_col_added or i == 0)
                        else last_drop_col
                    )
                    diffs.append(
                        {
                            "type": "add_col",
                            "source_struct_type": s,
                            "add_col_after": _add_col_after,
                            "parent_cols": parent_cols,
                        }
                    )
                    last_col_added = s.name

    # NOTE: Send all the nested schema change into a next level of struct
    for n in nested_schema_changes:
        _parent_cols: list[str] = copy.deepcopy(parent_cols)
        _parent_cols.append(n.get("table_struct_type").name)
        _func_out = extract_struct_type(
            source_struct=n.get("source_struct_type").dataType,
            table_struct=n.get("table_struct_type").dataType,
            parent_cols=_parent_cols,
        )
        diffs.extend(_func_out)

    return diffs


def clean_col_except_item_from_extract_array(
    columns: list[str] | str,
) -> list[str] | str:
    """Clean the list of columns or columns (str), remove the element (a flag
    from extracting array).
    """
    if not isinstance(columns, (list, str)):
        raise NotImplementedError("columns must be str or list of str only")

    is_str: bool = False
    if isinstance(columns, str):
        is_str: bool = True
        _cols: list[str] = columns.split(".")
    else:
        _cols = columns

    clean_columns = [c for c in _cols if c != ENUM_EXTRACT_ARRAY_TYPE]
    return ".".join(clean_columns) if is_str else clean_columns


def summarize_changes(
    changes: list[dict],
) -> tuple[ListDictAny, ListDictAny, ListDictAny]:
    """Get detail changes of alter columns, add columns, and drop columns"""
    alter_list: ListDictAny = []
    add_list: ListDictAny = []
    drop_list: ListDictAny = []
    for change in changes:
        change_type = change["type"]
        parent_cols = [
            col
            for col in change["parent_cols"]
            if col != ENUM_EXTRACT_ARRAY_TYPE
        ]
        share_dict = {
            "name": change["source_struct_type"].name,
            "data_type": change["source_struct_type"].dataType.simpleString(),
            "parent_cols": parent_cols,
        }
        match change_type:
            case "alter_col":
                alter_list.append(share_dict)
            case "add_col":
                add_list.append(share_dict)
            case "drop_col":
                drop_list.append(share_dict)

    return alter_list, add_list, drop_list



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import Field
from pydantic.functional_validators import model_validator
from typing_extensions import Self

from ....__types import DictData
from ....errors import ToolTransformError
from ....utils import join_newline, to_snake_case
from ...__abc import BaseEngine
from .__abc import BaseDuckDBTransform
from .__models import ColumnMap

if TYPE_CHECKING:
    from duckdb import DuckDBPyRelation as Relation

logger = logging.getLogger("jett")


class Sql(BaseDuckDBTransform):
    """SQL Transform model."""

    op: Literal["sql"]
    alias: str = "df"
    sql: str | None = None
    sql_file: str | None = None
    allow_group_transform: bool = False

    @model_validator(mode="after")
    def __check_sql(self) -> Self:
        """Check the necessary field of SQL use pass only SQL statement nor
        SQL filepath.
        """
        if not self.sql and not self.sql_file:
            raise ValueError("SQL and SQL file should not be empty together.")
        elif self.sql and self.sql_file:
            logger.warning(
                "If pass SQL statement and SQL file location together, it will "
                "use SQL statement first."
            )
        return self

    def apply(
        self,
        df: Relation,
        engine: DictData,
        **kwargs,
    ) -> Relation:
        """Apply to SQL Execution with priority transform.

        Args:
            df (Relation): A Relation instance.
            engine (DictData): Any submodule of BaseEngine.
        """
        import duckdb

        logger.info("Start SQL transform ...")
        if self.sql:
            sql: str = self.sql
        else:
            _engine: BaseEngine = engine["engine"]
            sql_file: Path = _engine.parent_dir / self.sql_file
            try:
                sql: str = sql_file.read_text()
            except FileNotFoundError:
                raise ToolTransformError(
                    f"SQL file does not exists {sql_file.resolve()}"
                ) from None
        logger.info(f"SQL Statement\n{sql}")
        return duckdb.sql(sql, alias="df")


class RenameSnakeCase(BaseDuckDBTransform):
    op: Literal["rename_snakecase"]
    allow_group_transform: bool = False

    def apply(self, df: Relation, engine: DictData, **kwargs) -> Relation:
        """Apply to Rename Columns to Snake case."""
        import duckdb

        old_cols = df.columns
        new_cols = [
            f'"{col_name}" AS {to_snake_case(col_name)}'
            for col_name in old_cols
        ]
        logger.info(f"Start convert to snakecase: \n{join_newline(new_cols)}")
        return duckdb.sql(f"SELECT {', '.join(new_cols)} FROM df")


class RenameColumns(BaseDuckDBTransform):
    """Rename Columns Transform model."""

    op: Literal["rename_columns"]
    columns: list[ColumnMap] = Field(description="A list of ColumnMap object.")
    allow_group_transform: bool = Field(default=False)

    def apply(self, df: Relation, engine: DictData, **kwargs) -> Relation:
        """Apply to Rename Column transform."""
        import duckdb

        stm: str = (
            f"SELECT * "
            f"RENAME ( {', '.join(c.gen_source() for c in self.columns)} ) "
            f"FROM df"
        )
        logger.info(f"Rename columns statement:\n{stm}")
        return duckdb.sql(query=stm, alias="df")


class DropColumns(BaseDuckDBTransform):
    op: Literal["drop_columns"]
    target_col: list[str]

    def apply(self, df: Relation, **kwargs) -> Relation: ...

    def apply_group(self, df: Relation, **kwargs) -> str: ...


class ExcludeColumns(BaseDuckDBTransform):
    """Exclude Columns"""

    op: Literal["exclude_columns"]
    columns: list[str]

    def apply(self, df: Relation, engine: DictData, **kwargs) -> Relation:
        """Apply to Exclude Columns"""
        import duckdb

        return duckdb.sql(
            f"SELECT * EXCLUDE ({', '.join(self.columns)}) FROM df"
        )



from typing import Any, Literal

from pydantic import Field

from ... import Result
from ...__types import DictData
from ...models import Context, MetricEngine, MetricTransform
from ..__abc import BaseEngine


class Dbt(BaseEngine):
    type: Literal["dbt"]
    profile: str = Field(description="A DBT profile YAML filepath.")
    project: str = Field(description="A DBT project YAML filepath.")

    def ping(self) -> bool:
        """Ping the DBT profile and project location."""

    def execute(
        self,
        context: Context,
        engine: DictData,
        metric: MetricEngine,
    ) -> Any: ...

    def set_engine_context(self, context: Context, **kwargs) -> DictData:
        return {"engine": self}

    def set_result(self, df: Any, context: Context) -> Result:
        return Result()

    def apply(
        self,
        df: Any,
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> Any:
        return df



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal

from pydantic import Field

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, SparkSession
    from pyspark.sql.connect.column import Column as ColumnRemote
    from pyspark.sql.connect.session import DataFrame as DataFrameRemote
    from pyspark.sql.types import StructType

    PairCol = tuple[Column, str]
    AnyDataFrame = DataFrame | DataFrameRemote
    AnyApplyGroupOutput = PairCol | list[PairCol]
    AnyApplyOutput = AnyApplyGroupOutput | AnyDataFrame

from ....__types import DictData
from ....errors import ToolTransformError
from ....models import (
    Context,
    MetricOperator,
    MetricOperatorOrder,
    MetricOperatorTransform,
)
from ....utils import sort_non_sensitive_str
from ...__abc import BaseTransform
from ..utils import extract_cols_without_array, schema2dict

logger = logging.getLogger("jett")


def is_pair_col(value: Any) -> bool:
    """Change value is a pair of Column and string.

    Args:
        value (Any): A pair of Column and its name or any value.

    Returns:
        bool: True if an input value is a pair of Column and alias.
    """
    return (
        isinstance(value, tuple)
        and len(value) == 2
        and isinstance(value[0], (Column, ColumnRemote))
        and isinstance(value[1], str)
    )


class BaseSparkTransform(BaseTransform, ABC):
    """Base Spark Transform abstract model."""

    priority: Literal["pre", "group", "post"] = Field(
        default="pre",
        description=(
            "A priority value for order transform operator before apply. "
            "This field will deprecate in the future and move to use `group` "
            "operator instead."
        ),
    )
    cache: bool = Field(
        default=False,
        description="Use `.cache` method to applied Spark DataFrame if it set.",
    )

    @abstractmethod
    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorOrder,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyApplyOutput:
        """Apply operator transform abstraction method.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.
            spark (SparkSession, default None): A Spark session.

        Returns:
            AnyApplyOutput: An any applied output that can be
                - PairCol: A pair of Column and its alias name.
                - list[PairCol]: A list of pair Column.
                - DataFrame: A Spark DataFrame.
        """

    def apply_group(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperator,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyApplyGroupOutput:
        """Apply group transform method that is optional apply for supported
        group transform model.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperator): A metric transform that was set from
                handler step for passing custom metric data.
            spark (SparkSession, default None): A Spark session.

        Returns:
            AnyApplyGroupOutput: An any applied group output that can be
                - PairCol: A pair of Column and its alias name.
                - list[PairCol]: A list of pair Column.
        """
        raise NotImplementedError(
            f"Transform: {self.__class__.__name__} on Spark engine does not "
            f"implement the group operation, please change this to priority "
            f"operator."
        )

    def handle_apply_group(
        self,
        df: DataFrame,
        context: Context,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> dict[str, Column]:
        """Handle Apply group transform.

        Args:
            df (AnyDataFrame): A Spark DataFrame.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            dict[str, Column]: A mapping of alias name and Column object.
        """
        logger.info(f"👷🔨 Handle Apply Group Operator: {self.op!r}")
        metric = MetricOperator(type="order", trans_op=self.op)
        context["metric_group_transform"].transforms.append(metric)
        try:
            output: PairCol | list[PairCol] = self.apply_group(
                df, engine, metric=metric, spark=spark, **kwargs
            )
            if is_pair_col(output):
                rs: dict[str, Column] = {output[1]: output[0]}
            elif isinstance(output, list) and all(
                is_pair_col(i) for i in output
            ):
                rs: dict[str, Column] = {p[1]: p[0] for p in output}
            else:
                raise ToolTransformError(
                    "Transform group on Spark engine should return the apply group "
                    "result with a pair of Column and str only"
                )
            if self.cache:
                logger.warning(
                    f"🏭 Cache the DataFrame after apply operator: {self.op!r}"
                )
                df.cache()
            return rs
        finally:
            self.post_apply(engine=engine, **kwargs)
            metric.finish()

    def handle_apply(
        self,
        df: AnyDataFrame,
        context: Context,
        engine: DictData,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyDataFrame:
        """Handle the Operator Apply result output from its custom apply that
        can make different type of result. It can be Column, DataFrame, or
        a pair of Column and its alias name.

        Args:
            df (AnyDataFrame): A Spark DataFrame.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            AnyDataFrame: A applied Spark DataFrame.
        """
        logger.info(f"👷🔧 Handle Apply Operator: {self.op!r}")
        metric = MetricOperatorOrder(type="order", trans_op=self.op)
        context["metric_operator"].append(metric)
        try:
            pre_schema: StructType = df.schema
            output: AnyApplyOutput = self.apply(
                df, engine=engine, metric=metric, **kwargs
            )
            if is_pair_col(output):
                df = df.withColumn(output[1], output[0])
            elif isinstance(output, list) and all(
                is_pair_col(i) for i in output
            ):
                df = df.withColumns({i[1]: i[0] for i in output})
            elif isinstance(output, (DataFrame, DataFrameRemote)):
                df: AnyDataFrame = output
            else:
                metric.add(key=self.op, value=output)
                logger.info(f"Set metric from func {self.op} completely")

                # ARCHIVE:
                # raise ToolTransformError(
                #     "Transform priority or fallback on Spark engine should return "
                #     "the apply result with type in a pair of Column, a list of "
                #     "pair of Column, or DataFrame only"
                # )
            if self.cache:
                logger.warning(
                    f"🏭 Cache the DataFrame after apply operator: {self.op!r}"
                )
                df.cache()

            # NOTE: Sync schema before return.
            self.sync_schema(pre_schema, df.schema, metric=metric, spark=spark)
            return df
        finally:
            self.post_apply(engine, **kwargs)
            metric.finish()

    @staticmethod
    def sync_schema(
        pre: StructType,
        post: StructType,
        *,
        metric: MetricOperatorTransform,
        spark: SparkSession | None = None,
    ) -> None:
        """Sync schema change to the metric transform.

        Args:
            pre (StructType): A pre schema before apply.
            post (StructType): A post schema that have applied.
            metric (MetricOperatorTransform): An operator transform metric model
                that want to update schema pre and post for tracking change.
            spark (SparkSession, default None): A Spark session.
        """
        pre_schema = spark.createDataFrame(data=[], schema=pre).schema
        post_schema = spark.createDataFrame(data=[], schema=post).schema
        pre_no_array = sort_non_sensitive_str(
            extract_cols_without_array(schema=pre_schema)
        )
        post_no_array = sort_non_sensitive_str(
            extract_cols_without_array(schema=post_schema)
        )
        # NOTE: Start update the pre- and post-schema metric.
        metric.transform_pre = {
            "schema": schema2dict(pre, sorted_by_name=True),
            "schema_no_array": pre_no_array,
        }
        metric.transform_post = {
            "schema": schema2dict(post, sorted_by_name=True),
            "schema_no_array": post_no_array,
        }



from typing import Any, Literal

from pyarrow import Table
from pyarrow.csv import ParseOptions, ReadOptions, read_csv
from pyarrow.dataset import CsvFileFormat, Dataset, dataset
from pyarrow.json import read_json
from pydantic import Field

from .....__types import DictData
from .....models import MetricSource, Shape
from ....__abc import BaseSource


class LocalJsonFileTable(BaseSource):
    type: Literal["local"]
    arrow_type: Literal["table"]
    file_format: Literal["json"]
    path: str

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Table, Shape]:
        table: Table = read_json(self.path)
        return table, Shape.from_tuple(table.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalCsvFileTable(BaseSource):
    type: Literal["local"]
    arrow_type: Literal["table"]
    file_format: Literal["csv"]
    path: str

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Table, Shape]:
        table: Table = read_csv(
            self.path,
            read_options=ReadOptions(
                autogenerate_column_names=True,
            ),
        )
        return table, Shape.from_tuple(table.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalCsvFileDataset(BaseSource):
    type: Literal["local"]
    arrow_type: Literal["dataset"]
    file_format: Literal["csv"]
    path: str
    partitioning: list[str] | str | None = Field(default=None)

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Dataset, Shape]:
        ds: Dataset = dataset(
            self.path,
            partitioning="hive",
            format=CsvFileFormat(
                **{"parse_options": ParseOptions(delimiter=",")},
            ),
        )
        table = ds.to_table()
        return table, Shape.from_tuple(table.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalParquetFileDataset(BaseSource):
    type: Literal["local"]
    arrow_type: Literal["dataset"]
    file_format: Literal["parquet"]
    path: str

    def load(
        self,
        engine: DictData,
        metric: MetricSource,
        **kwargs,
    ) -> tuple[Any, Shape]: ...

    def inlet(self) -> tuple[str, str]: ...
